@@ -500,7 +500,12 @@ The C<node> and <path> arguments are passed to as_node_path_list().
 The C<separator> argument is used to join the elemets of the path for each leaf node.
 
 The C<sortsub> argument is used to pass in a ref to a sub that will order the list.
-The subroutine will be passed a reference to the array returned by as_node_path_list().
+The subroutine will be passed a reference to the array returned by
+as_node_path_list() and should sort the contents of the array in place.
+The return value from the sub is ignored. For example, to sort the nodes by the
+second level key you could use:
+
+  sortsub => sub { my $ary=shift; @$ary = sort { $a->[2] cmp $b->[2] } @$ary }
 
 The C<format> argument is a C<sprintf> format string that specifies the format
 to use for each leaf node.  It uses the explicit format parameter index
@@ -508,11 +513,11 @@ mechanism to specify which of the arguments should appear where in the string.
 The arguments to sprintf are:
 
      1:  path to node, joined with the separator
-     2:  average duration (total/count)
+     2:  average duration (total duration/count)
          (3 thru 9 are currently unused)
     10:  count
     11:  total duration
-    12:  first_duration
+    12:  first duration
     13:  smallest duration
     14:  largest duration
     15:  time of first call
@@ -520,7 +525,7 @@ The arguments to sprintf are:
 
 =head1 CUSTOM DATA MANIPULATION
 
-Recall that C<$h->{Profile}->{Data}> is a reference to the collected data.
+Recall that C<< $h->{Profile}->{Data} >> is a reference to the collected data.
 Either to a 'leaf' array (when the Path is empty, i.e., DBI_PROFILE env var is 1),
 or a reference to hash containing values that are either further hash
 references or leaf array references.
@@ -607,11 +612,10 @@ default value for $method.
 The $h->{Profile}{Path} attribute is processed by dbi_profile() in
 the usual way.
 
-It is recommended that you keep these extra data samples separate
-from the DBI profile data samples by using values for $statement
-and $method that are distinct from any that are likely to appear
-in the profile data normally.
-
+The $h parameter is usually a DBI handle but it can also be a reference to a
+hash, in which case the dbi_profile() acts on each defined value in the hash.
+This is an efficient way to update multiple profiles with a single sample,
+and is used by the L<DashProfiler> module.
 
 =head1 SUBCLASSING
 
@@ -667,14 +671,14 @@ many forward references.  (Patches welcome!)
 
 
 use strict;
-use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION $ON_DESTROY_DUMP);
+use vars qw(@ISA @EXPORT @EXPORT_OK $VERSION);
 use Exporter ();
 use UNIVERSAL ();
 use Carp;
 
 use DBI qw(dbi_time dbi_profile dbi_profile_merge_nodes dbi_profile_merge);
 
-$VERSION = sprintf("2.%06d", q$Revision: 9647 $ =~ /(\d+)/o);
+$VERSION = sprintf("2.%06d", q$Revision: 13581 $ =~ /(\d+)/o);
 
 
 @ISA = qw(Exporter);
@@ -695,7 +699,8 @@ use constant DBIprofile_Statement	=> '!Statement';
 use constant DBIprofile_MethodName	=> '!MethodName';
 use constant DBIprofile_MethodClass	=> '!MethodClass';
 
-$ON_DESTROY_DUMP = sub { DBI->trace_msg(shift, 0) };
+our $ON_DESTROY_DUMP = sub { DBI->trace_msg(shift, 0) };
+our $ON_FLUSH_DUMP   = sub { DBI->trace_msg(shift, 0) };
 
 sub new {
     my $class = shift;
@@ -757,6 +762,25 @@ sub _auto_new {
 }
 
 
+sub empty {             # empty out profile data
+    my $self = shift;
+    DBI->trace_msg("profile data discarded\n",0) if $self->{Trace};
+    $self->{Data} = undef;
+}   
+
+sub filename {          # baseclass method, see DBI::ProfileDumper
+    return undef;
+}
+
+sub flush_to_disk {     # baseclass method, see DBI::ProfileDumper & DashProfiler::Core
+    my $self = shift;
+    return unless $ON_FLUSH_DUMP;
+    return unless $self->{Data};
+    my $detail = $self->format();
+    $ON_FLUSH_DUMP->($detail) if $detail;
+}
+
+
 sub as_node_path_list {
     my ($self, $node, $path) = @_;
     # convert the tree into an array of arrays
@@ -785,6 +809,8 @@ sub as_node_path_list {
 sub as_text {
     my ($self, $args_ref) = @_;
     my $separator = $args_ref->{separator} || " > ";
+    my $format_path_element = $args_ref->{format_path_element}
+        || "%s"; # or e.g., " key%2$d='%s'"
     my $format    = $args_ref->{format}
         || '%1$s: %11$fs / %10$d = %2$fs avg (first %12$fs, min %13$fs, max %14$fs)'."\n";
     
@@ -796,15 +822,19 @@ sub as_text {
     my $separator_re = eval($eval) || quotemeta($separator);
     #warn "[$eval] = [$separator_re]";
     my @text;
+    my @spare_slots = (undef) x 7;
     for my $node_path (@node_path_list) {
         my ($node, @path) = @$node_path;
-        s/[\r\n]+/ /g for @path;
-        s/$separator_re/ /g for @path;
-        my $path_str = join($separator, @path);
+        my $idx = 0;
+        for (@path) {
+            s/[\r\n]+/ /g;
+            s/$separator_re/ /g;
+            $_ = sprintf $format_path_element, $_, ++$idx;
+        }
         push @text, sprintf $format,
-            $path_str,                                # 1=path
-            ($node->[0] ? $node->[4]/$node->[0] : 0), # 2=avg
-            (undef) x 7,                              # spare slots
+            join($separator, @path),                  # 1=path
+            ($node->[0] ? $node->[1]/$node->[0] : 0), # 2=avg
+            @spare_slots,
             @$node; # 10=count, 11=dur, 12=first_dur, 13=min, 14=max, 15=first_called, 16=last_called
     }       
     return @text if wantarray;
